"""
Credo Governance
"""

import json
from typing import List, Optional, Union

from connect import __version__
from connect.evidence import Evidence, EvidenceRequirement
from connect.utils import check_subset, global_logger, json_dumps, wrap_list
from json_api_doc import deserialize, serialize

from .credo_api import CredoApi
from .credo_api_client import CredoApiClient


class Governance:
    """Class to store governance data.

    Governance is used to interact with the CredoAI Governance(Report) App.
    It has two main jobs.
    1. Get evidence_requirements of use_case and policy pack.
    2. Upload evidences gathered with evidence_requirements

    Parameters
    ----------
    credo_api_client: CredoApiClient, optional
        Credo API client. Uses default Credo API client if it is None
        Default Credo API client uses `~/.credo_config` to read API server configuration.
        Please prepare `~/.credo_config` file by downloading it from CredoAI Governance App.(My Settings > Tokens)

    Examples
    --------
    If you want to use your own configuration:

        from connect.governance.credo_api_client import CredoApiClient, CredoApiConfig
        from connect.governance.goverance import Governance

        config = CredoApiConfig(
            api_key="API_KEY", tenant="credo", api_server="https://api.credo.ai"
        )
        # or using credo_config file
        config = CredoApiConfig()
        config.load_config("CREDO_CONFIG_FILE_PATH")

        client = CredoApiClient(config=config)
        governace = Governance(credo_api_client=client)

    """

    def __init__(
        self, config_path: str = None, credo_api_client: CredoApiClient = None
    ):
        """Governance object to connect Lens with Credo AI Platform

        Parameters
        ----------
        config_path : str, optional
            path to .credoconfig file. If None, points to ~/.credoconfig, by default None
        credo_api_client : CredoApiClient, optional
            If provided, overrides the API configuration defined by
            the config path, by default None
        """
        self._use_case_id: Optional[str] = None
        self._policy_pack_id: Optional[str] = None
        self._evidence_requirements: List[EvidenceRequirement] = []
        self._evidences: List[Evidence] = []
        self._model = None
        self._plan: Optional[dict] = None
        self._unique_tags: List[dict] = []

        if credo_api_client:
            client = credo_api_client
        else:
            client = CredoApiClient(config_path=config_path)

        self._api = CredoApi(client=client)

    def register(
        self,
        assessment_plan_url: str = None,
        use_case_name: str = None,
        policy_pack_key: str = None,
        assessment_plan: str = None,
        assessment_plan_file: str = None,
    ):
        """
        Parameters
        ----------
        assessment_plan_url : str
            assessment plan URL
        use_case_name : str
            use case name
        policy_pack_key : str
            policy pack key
        assessment_plan : str
            assessment plan JSON string
        assessment_plan_file : str
            assessment plan file name that holds assessment plan JSON string

        Examples
        --------
        Get assessment plan and register it.
        There are three ways to do it:

            1. With assessment_plan_url.

                gov.register(assessment_plan_url="https://api.credo.ai/api/v2/tenant/use_cases/{id}/assessment_plans/{pp_id}")

            2. With use case name and policy pack key.

                gov.register(use_case_name="Fraud Detection", policy_pack_key="FAIR")

            3. With assessment_plan json string or filename. It is used in the air-gap condition.

                gov.register(assessment_plan="JSON_STRING")
                gov.register(assessment_plan_file="FILENAME")

        After successful registration, `gov.registered` returns True and able to get evidence_requirements:

            gov.registered    # returns True
            gov.get_evidence_requirements()


        """
        self._plan = None

        plan = None
        if use_case_name:
            assessment_plan_url = self._api.get_assessment_plan_url(
                use_case_name, policy_pack_key
            )

        if assessment_plan_url:
            plan = self._api.get_assessment_plan(assessment_plan_url)

        if assessment_plan:
            plan = self.__parse_json_api(assessment_plan)

        if assessment_plan_file:
            with open(assessment_plan_file, "r") as f:
                json_str = f.read()
                plan = self.__parse_json_api(json_str)

        if plan:
            self._plan = plan
            self._use_case_id = plan.get("use_case_id")
            self._policy_pack_id = plan.get("policy_pack_id")
            self._evidence_requirements = list(
                map(
                    lambda d: EvidenceRequirement(d),
                    plan.get("evidence_requirements", []),
                )
            )

            # Extract unique tags
            for x in self._evidence_requirements:
                if x.tags not in self._unique_tags:
                    self._unique_tags.append(x.tags)
            self._unique_tags = [x for x in self._unique_tags if x]

            global_logger.info(
                f"Successfully registered with {len(self._evidence_requirements)} evidence requirements"
            )

            if self._unique_tags:
                global_logger.info(
                    f"The following tags have being found in the evidence requirements: {self._unique_tags}"
                )

            self.clear_evidence()

    def __parse_json_api(self, json_str):
        return deserialize(json.loads(json_str))

    @property
    def registered(self):
        return bool(self._plan)

    def add_evidence(self, evidences: Union[Evidence, List[Evidence]]):
        """
        Add evidences
        """
        self._evidences += wrap_list(evidences)

    def clear_evidence(self):
        self.set_evidence([])

    def export(self, filename=None):
        """
        Upload evidences to CredoAI Governance(Report) App

        Returns
        -------
        True
            When uploading is successful with all evidence
        False
            When it is not registered yet, or evidence is insufficient
        """
        if not self._validate_export():
            return False
        to_return = self._match_requirements()

        if filename is None:
            self._api_export()
        else:
            self._file_export(filename)

        if to_return:
            export_status = "All requirements were matched."
        else:
            export_status = "Partial match of requirements."

        global_logger.info(export_status)
        return to_return

    def get_evidence_requirements(self, tags: dict = None):
        """
        Returns evidence requirements. Each evidence requirement can have optional tags
        (a dictionary). Only returns requirements that have tags that match the model
        (if provided), which are tags with the same tags as the model, or no tags.

        Parameters
        ----------
        tags : dict, optional
            Tags to subset evidence requirements. If a model has been set, will default
            to the model's tags. Evidence requirements will be returned that have no
            tags or have the same tag as provided.

        Returns
        -------
        List[EvidenceRequirement]
        """
        if tags is None:
            tags = self.get_model_tags()

        reqs = [
            e for e in self._evidence_requirements if (not e.tags or e.tags == tags)
        ]
        return reqs

    def get_evidence_tags(self):
        """Return the unique tags used for all evidence requirements"""
        return self._unique_tags

    def get_model_tags(self):
        """Get the tags for the associated model"""
        if self._model:
            return self._model["tags"]
        else:
            return None

    def set_artifacts(self, model, training_dataset=None, assessment_dataset=None):
        """Sets up internal knowledge of model and datasets to send to Credo AI Platform"""
        global_logger.info(
            f"Adding model ({model.name}) to governance. Model has tags: {model.tags}"
        )
        prepared_model = {"name": model.name, "tags": model.tags}
        if training_dataset:
            prepared_model["training_dataset_name"] = training_dataset.name
        if assessment_dataset:
            prepared_model["assessment_dataset_name"] = assessment_dataset.name
        self._model = prepared_model

    def set_evidence(self, evidences: List[Evidence]):
        """
        Update evidences
        """
        self._evidences = evidences

    def tag_model(self, model):
        """Interactive utility to tag a model tags from assessment plan"""
        tags = self.get_evidence_tags()
        print(f"Select tag from assessment plan to associated with model:")
        print("0: No tags")
        for number, tag in enumerate(tags):
            print(f"{number+1}: {tag}")
        selection = int(input("Number of tag to associate: "))
        if selection == 0:
            selected_tag = None
        else:
            selected_tag = tags[selection - 1]
        print(f"Selected tag = {selected_tag}. Applying to model...")
        model.tags = selected_tag
        if self._model:
            self._model["tags"] = selected_tag

    def _api_export(self):
        global_logger.info(
            f"Uploading {len(self._evidences)} evidences.. for use_case_id={self._use_case_id} policy_pack_id={self._policy_pack_id}"
        )
        self._api.create_assessment(self._use_case_id, self._prepare_export_data())

    def _check_inclusion(self, label, evidence):
        matching_evidence = []
        for e in evidence:
            if check_subset(label, e.label):
                matching_evidence.append(e)
        if not matching_evidence:
            return False
        if len(matching_evidence) > 1:
            global_logger.error(
                "Multiple evidence labels were found matching one requirement"
            )
            return False
        return matching_evidence

    def _file_export(self, filename):
        global_logger.info(
            f"Saving {len(self._evidences)} evidences to {filename}.. for use_case_id={self._use_case_id} policy_pack_id={self._policy_pack_id} "
        )
        data = self._prepare_export_data()
        meta = {"client": "Credo AI Lens", "version": __version__}
        data = json_dumps(serialize(data=data, meta=meta))
        with open(filename, "w") as f:
            f.write(data)

    def _match_requirements(self):
        missing = []
        required_labels = [e.label for e in self.get_evidence_requirements()]
        for label in required_labels:
            matching_evidence = self._check_inclusion(label, self._evidences)
            if not matching_evidence:
                missing.append(label)
                global_logger.info(f"Missing required evidence with label ({label}).")
            else:
                matching_evidence[0].label = label
        return not bool(missing)

    def _prepare_export_data(self):
        evidences = self._prepare_evidences()
        data = {
            "policy_pack_id": self._policy_pack_id,
            "models": [self._model] if self._model else None,
            "evidences": evidences,
            "$type": "assessments",
        }
        return data

    def _prepare_evidences(self):
        evidences = list(map(lambda e: e.struct(), self._evidences))
        return evidences

    def _validate_export(self):
        if not self.registered:
            global_logger.info("Governance is not registered, please register first")
            return False

        if 0 == len(self._evidences):
            global_logger.info(
                "No evidences added to governance, please add evidences first"
            )
            return False
        return True



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

import os
import sys

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import furo

sys.path.insert(0, os.path.abspath(".."))
import connect

# -- Project information -----------------------------------------------------

project = "Credo AI | Connect"
copyright = "2021, Credo AI Development Team"
author = "Credo AI Development Team"
release = connect.__version__

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.intersphinx",  # Interdocument references
    "sphinx.ext.autodoc",  # Core library for html generation from docstrings
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.autosummary",  # Create neat summary tables
    "sphinx_copybutton",
    "sphinx.ext.doctest",
    "sphinx.ext.coverage",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",  # needs to be AFTER napoleon
    "sphinx_rtd_theme",
    "nbsphinx",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# Auto-Doc Options
# ----------------
# Change the ordering of the member documentation
autodoc_member_order = "groupwise"
autoclass_content = "both"
autosummary_generate = True  # Turn on sphinx.ext.autosummary
# Enable 'expensive' imports for sphinx_autodoc_typehints
set_type_checking_flag = True
nbsphinx_allow_errors = True  # Continue through Jupyter errors
nbsphinx_execute = "never"  # do not execute jupyter notebooks

autodoc_mock_imports = [
    "dotenv",
    "json_api_doc",
    "numpy",
    "pandas",
]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#


# documentation for furo: https://pradyunsg.me/furo/quickstart/

html_theme = "furo"

# import sphinx_rtd_theme
# html_theme = 'sphinx_rtd_theme'
# html_theme_options = {
#    'navigation_depth': 6
# }
html_static_path = ["_static"]
html_css_files = [
    "css/custom.css",
]



"""
Test Credo API client
"""

import os
import pathlib
import pytest
import responses
from credoai import __version__
from credoai.governance.credo_api_client import CredoApiClient, CredoApiConfig


class TestCredoApiConfig:
    def test_credo_api_config(self):
        config = CredoApiConfig(
            api_key="API_KEY", tenant="credo", api_server="http://localhost:4000"
        )

        assert "API_KEY" == config.api_key
        assert "credo" == config.tenant
        assert "http://localhost:4000" == config.api_server
        assert "http://localhost:4000/api/v2/credo" == config.api_base

    def test_default_config_path(self):
        path = CredoApiConfig.default_config_path()
        assert path.endswith(".credoconfig")

    def test_load_config(self):
        # Assumes that .credoconfig file exists in the same directory
        path = pathlib.Path(__file__).parent.resolve()
        config_path = os.path.join(path, ".credoconfig")

        config = CredoApiConfig()
        config.load_config(config_path)

        assert "AbcdeF" == config.api_key
        assert "credoai" == config.tenant
        assert "http://localhost:4000" == config.api_server
        assert "http://localhost:4000/api/v2/credoai" == config.api_base

    def test_valid(self):
        config = CredoApiConfig()
        assert False == config.valid

        config.load_config("/not_exist/file")
        assert False == config.valid

        config = CredoApiConfig(
            api_key="API_KEY", tenant="credo", api_server="http://localhost:4000"
        )
        assert True == config.valid


API_KEY = "API_KEY"
API_SERVER = "http://api.server"
TENANT = "credoai"


class TestCredoApiClient:
    # Tests use mocking with responses lib(https://pypi.org/project/responses/) instead of sending API request to the server

    @pytest.fixture()
    @responses.activate
    def client(self):
        responses.post(
            f"{API_SERVER}/auth/exchange", json={"access_token": "VALID_TOKEN"},
        )

        config = CredoApiConfig(api_key=API_KEY, api_server=API_SERVER, tenant=TENANT)
        return CredoApiClient(config=config)

    @responses.activate
    def test_refresh_token(self, client):
        responses.post(
            f"{API_SERVER}/auth/exchange",
            json={"access_token": "REFRESHED_VALID_TOKEN"},
        )

        client.refresh_token()

        headers = client._session.headers
        assert "Bearer REFRESHED_VALID_TOKEN" == headers.get("Authorization")
        assert "application/vnd.api+json" == headers.get("content-type")
        assert "Credo AI Lens" == headers.get("X-Client-Name")
        assert __version__ == headers.get("X-Client-Version")

    @responses.activate
    def test_refresh_token_with_invalid_config(self):
        responses.post(
            f"{API_SERVER}/auth/exchange",
            json={"access_token": "REFRESHED_VALID_TOKEN"},
        )

        invalid_config = CredoApiConfig()
        client = CredoApiClient(config=invalid_config)
        client.refresh_token()

        # it fails to get refresh token, so it does not update Authorization header
        headers = client._session.headers
        assert None == headers.get("Authorization")

    @responses.activate
    def test_access_token_expired(self, client):
        # return 401 with invalid token
        responses.get(
            f"{API_SERVER}/api/v2/{TENANT}/models",
            headers={"Authorization": "Bearer INVALID_TOKEN"},
            status=401,
        )

        # refresh token to REFRESHED_VALID_TOKEN
        responses.post(
            f"{API_SERVER}/auth/exchange",
            json={"access_token": "REFRESHED_VALID_TOKEN"},
        )

        # response 200 with REFRESHED_VALID_TOKEN
        responses.get(
            f"{API_SERVER}/api/v2/{TENANT}/models",
            headers={"Authorization": "Bearer REFRESHED_VALID_TOKEN"},
            json={
                "data": [
                    {"attributes": {"name": "model 1"}, "type": "models", "id": "123"}
                ]
            },
        )

        client.set_access_token("INVALID_TOKEN")
        # Client will try to send request with INVALID_TOKEN, and revices 401.
        # Then it tries refresh_token
        # Then, try the request again and succeed.
        # As a result of it, the Authorization header is updated with REFRESHED_VALID_TOKEN
        response = client.get("models")
        headers = client._session.headers
        assert "Bearer REFRESHED_VALID_TOKEN" == headers.get("Authorization")
        assert 1 == len(response)

    @responses.activate
    def test_get_request(self, client):
        responses.get(
            f"{API_SERVER}/api/v2/{TENANT}/models",
            json={
                "data": [
                    {"attributes": {"name": "model 1"}, "type": "models", "id": "123"}
                ]
            },
        )

        response = client.get("models")
        assert 1 == len(response)
        model = response[0]
        assert "model 1" == model["name"]

    @responses.activate
    def test_post_request(self, client):
        responses.post(
            f"{API_SERVER}/api/v2/{TENANT}/models",
            json={
                "data": {
                    "attributes": {"name": "model 1"},
                    "type": "models",
                    "id": "123",
                }
            },
        )

        response = client.post("models", {"name": "model 1", "$type": "models"})
        assert "model 1" == response["name"]

    @responses.activate
    def test_patch_request(self, client):
        responses.patch(
            f"{API_SERVER}/api/v2/{TENANT}/models",
            json={
                "data": {
                    "attributes": {"name": "model 1"},
                    "type": "models",
                    "id": "123",
                }
            },
        )

        response = client.patch(
            "models", {"name": "model 1", "$type": "models", "id": "123"}
        )
        assert "model 1" == response["name"]

    @responses.activate
    def test_delete_request(self, client):
        responses.delete(f"{API_SERVER}/api/v2/{TENANT}/models/123",)

        response = client.delete("models/123")
        assert None == response



"""
Generic containers for evaluator results

The containers accept raw data from the evaluators and convert it into
suitable evidences.
"""
from abc import ABC, abstractmethod

import pandas as pd

from connect.utils import ValidationError

from .evidence import MetricEvidence, TableEvidence


class EvidenceContainer(ABC):
    def __init__(self, evidence_class, data, labels=None, metadata=None):
        """Abstract Class defining Evidence Containers

        Evidence Containers are light wrappers around data objects that
        validate their format for the purpose of evidence export. They
        define a "to_evidence" function which transforms the
        data into a particular evidence format

        Parameters
        ----------
        evidence_class : Evidence
            An Evidence class
        data :
            The data, formatted appropriately for the evidence type
            Typing is variable depending on evidence being wrapped
                Most metrics will be pd.DataFrames
                Evidence generated by integrated packages (e.g. deepchecks) is dependent on
                package's output
        labels : dict
            Additional labels to pass to underlying evidence
        metadata : dict
            Metadata to pass to underlying evidence
        """
        self.evidence_class = evidence_class
        self._validate_inputs(data)
        self._validate(data)
        self._data = data
        self.labels = labels
        self.metadata = metadata or {}

    @property
    def data(self):
        return self._data

    @abstractmethod
    def to_evidence(self):
        pass

    def _validate_inputs(self, data):
        """Default validation assumes data is a dataframe"""
        if not isinstance(data, pd.DataFrame):
            raise ValidationError("'data' must be a dataframe")

    @abstractmethod
    def _validate(self, data):
        pass


class MetricContainer(EvidenceContainer):
    """Containers for all Metric type evidence"""

    def __init__(self, data: pd.DataFrame, labels: dict = None, metadata: dict = None):
        super().__init__(MetricEvidence, data, labels, metadata)

    def to_evidence(self, **metadata):
        evidence = []
        for _, data in self._data.iterrows():
            evidence.append(
                self.evidence_class(
                    additional_labels=self.labels, **data, **self.metadata, **metadata
                )
            )
        return evidence

    def _validate(self, data):
        required_columns = {"type", "value"}
        column_overlap = data.columns.intersection(required_columns)
        if len(column_overlap) != len(required_columns):
            raise ValidationError(f"Must have columns: {required_columns}")


class TableContainer(EvidenceContainer):
    """Container for all Table type evidence"""

    def __init__(self, data: pd.DataFrame, labels: dict = None, metadata: dict = None):
        super().__init__(TableEvidence, data, labels, metadata)

    def to_evidence(self, **metadata):
        return [
            self.evidence_class(
                self._data.name, self._data, self.labels, **self.metadata, **metadata
            )
        ]

    def _validate(self, data):
        try:
            data.name
        except AttributeError:
            raise ValidationError("DataFrame must have a 'name' attribute")



import collections
import hashlib
import json
from pathlib import Path
from typing import Any, Dict

import numpy as np
import pandas as pd


class NotRunError(Exception):
    pass


class ValidationError(Exception):
    pass


class InstallationError(Exception):
    pass


class IntegrationError(Exception):
    pass


class SupressSettingWithCopyWarning:
    def __enter__(self):
        pd.options.mode.chained_assignment = None

    def __exit__(self, *args):
        pd.options.mode.chained_assignment = "warn"


def check_subset(subset, superset):
    """Check whether one dictionary, list or set is a subset of another

    Handles nested dictionaries
    """
    if type(subset) != type(superset):
        return False
    if isinstance(subset, dict):
        for k, v in subset.items():
            superset_value = superset.get(k)
            if superset_value == v:
                continue
            elif type(v) != type(superset_value):
                return False
            elif isinstance(v, (dict, list, set)):
                out = check_subset(v, superset_value)
                if not out:
                    return False
            else:
                return False
    if isinstance(subset, (list, set)):
        return set(subset) <= set(superset)

    return True


def get_project_root() -> Path:
    return Path(__file__).parent.parent


def flatten_list(lst):
    return [item for sublist in lst for item in sublist]


def update_dictionary(d, u):
    """Recursively updates a dictionary"""
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = update_dictionary(d.get(k, {}), v)
        elif isinstance(v, list):
            d[k] = v + d.get(k, [])
        else:
            d[k] = v
    return d


def wrap_list(obj):
    """Ensures object is an iterable"""
    if type(obj) == str:
        obj = [obj]
    elif obj is None:
        return None
    try:
        iter(obj)
    except TypeError:
        obj = [obj]
    return obj


def remove_suffix(text, suffix):
    return text[: -len(suffix)] if text.endswith(suffix) and len(suffix) != 0 else text


def humanize_label(s):
    return " ".join(s.split("_")).title()


class CredoEncoder(json.JSONEncoder):
    """Special json encoder for numpy types"""

    def default(self, obj):
        # numpy encoders
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def json_dumps(obj):
    """Custom json dumps with encoder"""
    return json.dumps(obj, cls=CredoEncoder, indent=2)


def dict_hash(dictionary: Dict[str, Any]) -> str:
    """MD5 hash of a dictionary."""
    dhash = hashlib.md5()
    # We need to sort arguments so {'a': 1, 'b': 2} is
    # the same as {'b': 2, 'a': 1}
    encoded = json.dumps(dictionary, sort_keys=True).encode()
    dhash.update(encoded)
    return dhash.hexdigest()


def to_array(lst):
    """
    Converts list-like object to array
    Parameters
    ----------
    lst :  (List, pandas.Series, numpy.ndarray)
        The list-like to be converted
    """
    if type(lst) == pd.Series:
        return lst.values
    elif type(lst) == list:
        return np.array(lst)
    elif type(lst) == np.ndarray:
        return lst
    else:
        raise TypeError


def is_categorical(series, threshold=0.05):
    """Identifies whether a series is categorical or not

    Logic: If there are relatively few unique values for a feature, the feature is likely categorical.
    The results are estimates and are not guaranteed to be correct.

    Parameters
    ----------
    series : pd.Series
        Series to evaluate
    threshold : float
        The threshold (number of the unique values over the total number of values)


    Returns
    -------
    bool
        Whether the series is categorical or not
    """

    if series.dtype.name in ["category", "object"]:
        return True
    # float columns are assumed not-categorical
    elif len(series.unique()) / len(series) < threshold:
        return True
    else:
        return False



from connect.evidence import EvidenceContainer
from connect.utils import ValidationError

from .evidence import DataProfilerEvidence, ModelProfilerEvidence
from .utils import get_pandas_profile_type

PandasProfileType = get_pandas_profile_type()


class DataProfilerContainer(EvidenceContainer):
    """Container for all profiler type evidence"""

    def __init__(self, data, labels: dict = None, metadata: dict = None):
        super().__init__(DataProfilerEvidence, data, labels, metadata)

    def to_evidence(self, **metadata):
        return [
            self.evidence_class(
                self._data.get_description(), self.labels, **self.metadata, **metadata
            )
        ]

    def _validate(self, data):
        pass

    def _validate_inputs(self, data):
        if not isinstance(data, PandasProfileType):
            raise ValidationError(
                "'data' must be a pandas_profiling.profile_report.ProfileReport"
            )


class ModelProfilerContainer(EvidenceContainer):
    """Container for Model Profiler type evidence"""

    def __init__(self, data, labels=None, metadata=None):
        super().__init__(ModelProfilerEvidence, data, labels, metadata)

    def to_evidence(self, **metadata):
        return [
            self.evidence_class(self._data, self.labels, **self.metadata, **metadata)
        ]

    def _validate(self, data):
        necessary_index = ["parameters", "feature_names", "model_name"]
        if list(data.columns) != ["results"]:
            raise ValidationError(
                "Model profiler data must only have one column: 'results'"
            )
        if sum(data.index.isin(necessary_index)) != 3:
            raise ValidationError(f"Model profiler data must contain {necessary_index}")



from abc import ABC
from typing import List


class EvidenceRequirement(ABC):
    def __init__(
        self,
        data: dict,
    ):
        self._evidence_type: str = data.get("evidence_type")
        self._label: dict = data.get("label")
        self._sensitive_features: List[str] = data.get("sensitive_features", [])
        self._tags: dict = data.get("tags")

    def __str__(self) -> str:
        return f"{self.evidence_type}-EvidenceRequirement.label-{self.label}"

    @property
    def evidence_type(self):
        return self._evidence_type

    @property
    def label(self):
        return self._label

    @property
    def tags(self):
        return self._tags



"""
Everything needed to generate inputs for the Credo AI Platform

All the logic defined here is in order to standardize communication with 
Credo AI Platform API
"""
from .containers import EvidenceContainer, MetricContainer, TableContainer
from .evidence import Evidence, MetricEvidence, TableEvidence
from .evidence_requirement import EvidenceRequirement



import pytest
from credoai.governance.credo_api import CredoApi
from credoai.governance.credo_api_client import CredoApiClient


class TestCredoApi:
    @pytest.fixture()
    def client(self, mocker):
        # Use mock for CredoApiClient becasue we do not want to call HTTP request from test.
        # We will only check if CredoApi is calling client method with right arguments.
        mocker.patch.object(CredoApiClient, "get")
        mocker.patch.object(CredoApiClient, "post")
        mocker.patch.object(CredoApiClient, "patch")
        mocker.patch.object(CredoApiClient, "delete")
        mocker.patch.object(CredoApiClient, "refresh_token")

        return CredoApiClient()

    @pytest.fixture()
    def api(self, client):
        return CredoApi(client=client)

    # def test_real(self):
    #     client = CredoApiClient()
    #     api = CredoApi(client=client)

    #     response = api.get_assessment_plan(
    #         "http://localhost:4000/api/v2/credoai/use_cases/64YUaLWSviHgibJaRWr3ZE/assessment_plans/NYCE+1"
    #     )
    #     print(response)

    #     response = api.get_assessment_plan_url("Keyboard Sensitivity", "NYCE")
    #     print(response)

    def test_get_assessment_plan_url(self, api, client):
        api.get_assessment_plan_url("Fraud Detection", "FAIR")

        client.get.assert_called_with(
            "assessment_plan_url?use_case_name=Fraud Detection&policy_pack_key=FAIR"
        )

    def test_get_assessment_plan(self, api, client):
        url = "http://localhost:4000/api/v2/credoai/use_cases/64YUaLWSviHgibJaRWr3ZE/assessment_plans/NYCE+1"
        api.get_assessment_plan(url)

        client.get.assert_called_with(url)

    def test_create_assessment(self, api, client):
        use_case_id = "64YUaLWSviHgibJaRWr3ZE"
        policy_pack_id = "FAIR+1"

        evidences = [
            {
                "type": "metric",
                "label": {"metric_type": "accuracy_score"},
                "metadata": {
                    "model_name": "MODEL NAME 1",
                    "dataset_name": "DATASET NAME 1",
                },
                "data": {
                    "value": 0.6,
                    "confidence_interval_range": [0.1, 0.5],
                    "confidence_interval": 0.2,
                },
                "generated_at": "2022-05-03T11:33:25.582138Z",
            }
        ]
        body = {
            "policy_pack_id": policy_pack_id,
            "evidences": evidences,
            "$type": "assessments",
        }
        api.create_assessment(use_case_id, body)

        client.post.assert_called_with(f"use_cases/{use_case_id}/assessments", body)



import json
import tempfile

import pytest
from connect.evidence.evidence import MetricEvidence, TableEvidence
from connect.governance.credo_api import CredoApi
from connect.governance.credo_api_client import CredoApiClient
from connect.governance.governance import Governance
from pandas import DataFrame

USE_CASE_ID = "64YUaLWSviHgibJaRWr3ZE"
POLICY_PACK_ID = "NYCE+1"
EVIDENCE_REQUIREMENTS = [
    {"evidence_type": "metric", "label": {"metric_type": "accuracy_score"}},
    {"evidence_type": "metric", "label": {"metric_type": "p_value"}},
    {
        "evidence_type": "table",
        "label": {"table_name": "disaggregated_performance"},
        "sensitive_features": ["profession", "gender"],
    },
]
ASSESSMENT_PLAN_URL = f"http://api.credo.ai/api/v2/credoai/use_cases/${USE_CASE_ID}/assessment_plans/{POLICY_PACK_ID}"
ASSESSMENT_PLAN = {
    "use_case_id": USE_CASE_ID,
    "policy_pack_id": POLICY_PACK_ID,
    "evidence_requirements": EVIDENCE_REQUIREMENTS,
}

ASSESSMENT_PLAN_JSON_STR = json.dumps(
    {
        "data": {"attributes": ASSESSMENT_PLAN},
        "id": "id",
        "type": "assessment_plan",
    }
)


def build_metric_evidence(type):
    return MetricEvidence(
        type=type,
        value=0.2,
        model_name="superich detector",
        dataset_name="account data",
    )


def build_table_evidence(table_name):
    return TableEvidence(
        name=table_name, table_data=DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]})
    )


class TestGovernance:
    @pytest.fixture()
    def client(self, mocker):
        # Mocking client to not send actual request to the server
        mocker.patch.object(CredoApiClient, "get")
        mocker.patch.object(CredoApiClient, "post")
        mocker.patch.object(CredoApiClient, "patch")
        mocker.patch.object(CredoApiClient, "delete")
        mocker.patch.object(CredoApiClient, "refresh_token")

        return CredoApiClient()

    @pytest.fixture()
    def api(self, mocker):
        # Mocking api to simulate result
        mocker.patch.object(CredoApi, "get_assessment_plan")
        mocker.patch.object(CredoApi, "get_assessment_plan_url")
        mocker.patch.object(CredoApi, "create_assessment")

        return CredoApi()

    @pytest.fixture()
    def gov(self, client):
        return Governance(credo_api_client=client)

    @pytest.fixture
    def plan_file_mock(self, mocker):
        data = mocker.mock_open(read_data=ASSESSMENT_PLAN_JSON_STR)
        mocker.patch("builtins.open", data)

    # def test_real(self):
    #     client = CredoApiClient()
    #     gov = Governance(credo_api_client=client)

    #     gov.register(
    #         assessment_plan_url="http://localhost:4000/api/v2/credoai/use_cases/64YUaLWSviHgibJaRWr3ZE/assessment_plans/NYCE+1"
    #     )
    #     print(gov._use_case_id)

    def test_register_with_assessment_plan_url(self, gov, api):
        # mocking api result
        api.get_assessment_plan.return_value = ASSESSMENT_PLAN

        gov.register(assessment_plan_url=ASSESSMENT_PLAN_URL)

        api.get_assessment_plan.assert_called_with(ASSESSMENT_PLAN_URL)

        assert USE_CASE_ID == gov._use_case_id
        assert POLICY_PACK_ID == gov._policy_pack_id
        assert 3 == len(gov.get_evidence_requirements())
        assert True == gov.registered

    def test_register_with_use_case_name_and_policy_pack_key(self, gov, api):
        # mocking api result
        api.get_assessment_plan_url.return_value = ASSESSMENT_PLAN_URL
        api.get_assessment_plan.return_value = ASSESSMENT_PLAN

        gov.register(use_case_name="Fraud Detection", policy_pack_key="FAIR+1")

        api.get_assessment_plan_url.assert_called_with("Fraud Detection", "FAIR+1")
        api.get_assessment_plan.assert_called_with(ASSESSMENT_PLAN_URL)

        assert USE_CASE_ID == gov._use_case_id
        assert POLICY_PACK_ID == gov._policy_pack_id
        assert 3 == len(gov.get_evidence_requirements())
        assert True == gov.registered

    def test_register_with_use_case_name_and_policy_pack_key_not_exist(self, gov, api):
        api.get_assessment_plan_url.return_value = None

        gov.register(use_case_name="Fraud Detection", policy_pack_key="FAIR+1")

        assert False == gov.registered

    def test_register_with_assessment_plan(self, gov):

        gov.register(assessment_plan=ASSESSMENT_PLAN_JSON_STR)

        assert USE_CASE_ID == gov._use_case_id
        assert POLICY_PACK_ID == gov._policy_pack_id
        assert True == gov.registered

        req = gov._evidence_requirements[0]
        assert "metric" == req.evidence_type
        assert {"metric_type": "accuracy_score"} == req.label

    def test_register_with_assessment_plan_file(self, gov, plan_file_mock):

        gov.register(assessment_plan_file="filename")

        assert USE_CASE_ID == gov._use_case_id
        assert POLICY_PACK_ID == gov._policy_pack_id
        assert True == gov.registered

        req = gov._evidence_requirements[2]
        assert "table" == req.evidence_type
        assert {"table_name": "disaggregated_performance"} == req.label

    def test_add_evidences(self, gov):
        gov.set_evidence([build_metric_evidence("recall")])
        gov.add_evidence([build_metric_evidence("precision")])

        assert 2 == len(gov._evidences)

        gov.set_evidence([])
        assert 0 == len(gov._evidences)

    def test_export_without_registeration(self, gov):
        assert False == gov.export()

    def test_export_without_evidences(self, gov):
        gov.register(assessment_plan=ASSESSMENT_PLAN_JSON_STR)
        assert False == gov.export()

    def test_export_all_requirements_fulfilled(self, gov):
        gov.register(assessment_plan=ASSESSMENT_PLAN_JSON_STR)
        evidences = [
            build_metric_evidence("accuracy_score"),
            build_metric_evidence("p_value"),
            build_table_evidence("disaggregated_performance"),
        ]
        gov.add_evidence(evidences)
        assert True == gov.export()

    def test_export_partial_requirements_fulfilled(self, gov):
        gov.register(assessment_plan=ASSESSMENT_PLAN_JSON_STR)
        evidences = [
            build_metric_evidence("accuracy_score"),
        ]
        gov.add_evidence(evidences)
        assert False == gov.export()

    def test_export_to_file_all_requirements_fulfilled(self, gov):
        gov.register(assessment_plan=ASSESSMENT_PLAN_JSON_STR)
        evidences = [
            build_metric_evidence("accuracy_score"),
            build_metric_evidence("p_value"),
            build_table_evidence("disaggregated_performance"),
        ]
        gov.add_evidence(evidences)
        with tempfile.TemporaryDirectory() as tempDir:
            filename = f"{tempDir}/assessment.json"
            assert True == gov.export(filename)

    def test_export_to_file_partial_requirements_fulfilled(self, gov):
        gov.register(assessment_plan=ASSESSMENT_PLAN_JSON_STR)
        evidences = [
            build_metric_evidence("accuracy_score"),
        ]
        gov.add_evidence(evidences)
        with tempfile.TemporaryDirectory() as tempDir:
            filename = f"{tempDir}/assessment.json"
            assert False == gov.export(filename)



"""
Credo API functions
"""
from typing import List

from connect.utils import global_logger
from requests.exceptions import HTTPError

from .credo_api_client import CredoApiClient


class CredoApi:
    """
    CredoApi holds Credo API functions
    """

    def __init__(self, client: CredoApiClient = None):
        self._client = client

    def set_client(self, client: CredoApiClient):
        """
        Sets Credo Api Client

        Parameters
        ----------
        client : CredoApiClient
            Credo API client
        """
        self._client = client

    def get_assessment_plan_url(self, use_case_name: str, policy_pack_key: str = None):
        """
        Convert use_case_name and policy_pack_key to assessment_plan_url

        Parameters
        ----------
        use_case_name : str
            name of a use case
        policy_pack_key : str
            policy pack key, ie: FAIR
            If it is None, it gets the first resgisterd policy pack in use case

        Returns
        -------
        None
            When use_case_name does not exist or policy_pack_key is not registered to the use_case
        str
            assessment_plan_url

        Raises
        ------
        HTTPError
            When API request returns error other than 404
        """

        try:
            path = f"assessment_plan_url?use_case_name={use_case_name}"
            if policy_pack_key:
                path += f"&policy_pack_key={policy_pack_key}"
            response = self._client.get(path)
            return response["url"]
        except HTTPError as error:
            if policy_pack_key is not None:
                global_logger.info(
                    f"Use case ({use_case_name}) with policy pack ({policy_pack_key}) does not exist"
                )
            else:
                global_logger.info(
                    f"Cannot find assessment plan URL of use case {use_case_name}"
                )
            data = error.response.json()
            errors = data.get("errors", None)
            if errors:
                detail = errors[0]["detail"]
                if error:
                    global_logger.info(f"Error : {detail}")
                else:
                    raise error

                return None
            else:
                raise error

    def get_assessment_plan(self, url: str):
        """
        Get assessment plan from API server and returns it.

        Parameters
        ----------
        url : str
            assessment plan URL

        Returns
        -------
        dict
            evidence_requirements(list): list of evidence requirements
            policy_pack_id(str): policy pack id(key+version), ie: FAIR+1
            use_case_id(str): use case id

        Raises
        ------
        HTTPError
            When API request returns error
        """

        return self._client.get(url)

    def create_assessment(self, use_case_id: str, data: dict):
        """
        Upload evidences to API server.

        Parameters
        ----------
        use_case_id : str
            use case id
        data : dict
            assessment data generated by Governance

        Raises
        ------
        HTTPError
            When API request returns error
        """

        path = f"use_cases/{use_case_id}/assessments"
        return self._client.post(path, data)



#
from typing import TYPE_CHECKING, TypeVar


def get_deepchecks_type():
    """check if deepchecks is imported. If not, define object for type checking"""
    if TYPE_CHECKING:
        try:
            from deepchecks.core import SuiteResult

            deep_checks_type = SuiteResult
        except ImportError:
            deep_checks_type = TypeVar("SuiteResult")
    else:
        from deepchecks.core import SuiteResult

        deep_checks_type = SuiteResult
    return deep_checks_type



from connect.evidence import Evidence

from .utils import get_deepchecks_type

DeepChecksType = get_deepchecks_type()


class DeepchecksEvidence(Evidence):
    """
    Evidence for deepchecks SuiteResult objects

    Underlying evidence is passed as a JSON-formatted string. Other

    Parameters
    ----------
    name: string
        Name of the suite of deepchecks checks run by the Deepchecks evaluator
    result : SuiteResult object
        The output of a call to a deepchecks.Suite object's run() function.
    additional_labels : dict, optional
        Additional labels to be passed to, and displayed by the Credo AI Governance platform
    metadata : dict, optional
        Arbitrary keyword arguments to append to metric as metadata. These will be
        displayed in the governance app
    """

    def __init__(
        self,
        name: str,
        result: DeepChecksType,
        additional_labels: dict = None,
        **metadata,
    ):
        self.name = name
        self._result = result
        super().__init__("deepchecks_result", additional_labels, **metadata)

    @property
    def data(self):
        return self._result.to_json()

    @property
    def base_label(self):
        label = {"suite_result_name": self.name}
        return label



"""
Defines Credo API client
"""

import os
from typing import Dict

import requests
from connect import __version__
from connect.utils import global_logger, json_dumps
from dotenv import dotenv_values
from json_api_doc import deserialize, serialize

CREDO_URL = "https://api.credo.ai"


class CredoApiConfig:
    """
    Defines Credo API configs
    """

    def __init__(self, api_key: str = None, tenant: str = None, api_server: str = None):
        self._api_key = api_key
        self._tenant = tenant
        self._api_server = api_server
        self._api_base = self.__build_api_base()
        self.logger = global_logger

    @staticmethod
    def default_config_path():
        """
        Returns default config path
        """
        return os.path.join(os.path.expanduser("~"), ".credoconfig")

    @property
    def api_key(self):
        """
        Returns api_key
        """
        return self._api_key

    @property
    def api_server(self):
        """
        Returns api_server
        """
        return self._api_server

    @property
    def tenant(self):
        """
        Returns tenant
        """
        return self._tenant

    @property
    def api_base(self):
        """
        Returns api_base, which is base url for API request
        """
        return self._api_base

    @property
    def valid(self):
        return bool(self._api_key) and bool(self._api_server)

    def load_config(self, config_path=None):
        """
        Load API configs from the config_path
        """
        config_path = config_path or self.default_config_path()

        if not os.path.exists(config_path):
            self.logger.error("Path to config file Not Found")
            # return example
            self._api_key = None
            self._tenant = None
            self._api_server = None
        else:
            config = dotenv_values(config_path)
            self._tenant = config["TENANT"]
            self._api_key = config["API_KEY"]
            self._api_server = self.__build_api_server(config)
            self._api_base = self.__build_api_base()

    def __build_api_server(self, config):
        if config.get("API_URL"):
            return config["API_URL"].replace("/api/v2", "")

        return os.path.join(config.get("CREDO_URL", CREDO_URL))

    def __build_api_base(self):
        if self._api_server:
            return os.path.join(self._api_server, "api/v2", self._tenant)
        return None


class CredoApiClient:
    """
    CredoApiClient is interface class to the Credo API server.
    """

    def __init__(self, config: CredoApiConfig = None, config_path=None):
        if config:
            self._config = config
        else:
            self._config = CredoApiConfig()
            self._config.load_config(config_path=config_path)

        self._session = requests.Session()
        self.refresh_token()

    def refresh_token(self):
        """
        Get access token and set to headers
        """
        if self._config.valid:
            data = {"api_token": self._config.api_key, "tenant": self._config.tenant}
            headers = {"content-type": "application/json", "charset": "utf-8"}
            auth_url = os.path.join(self._config.api_server, "auth", "exchange")
            response = requests.post(auth_url, json=data, headers=headers)
            access_token = response.json()["access_token"]
            self.set_access_token(access_token)

    def set_access_token(self, access_token):
        """
        Set access token to headers
        """
        headers = {
            "Authorization": f"Bearer {access_token}",
            "accept": "application/vnd.api+json",
            "content-type": "application/vnd.api+json",
            "X-Client-Name": "Credo AI Lens",
            "X-Client-Version": __version__,
        }
        self._session.headers.update(headers)

    def __make_request(self, method: str, path: str, **kwargs):

        if path.startswith("http"):
            endpoint = path
        else:
            endpoint = self.__build_endpoint(path)

        response = self._session.request(method, endpoint, **kwargs)
        if response.status_code == 401:
            self.refresh_token()
            response = self._session.request(method, endpoint, **kwargs)

        response.raise_for_status()

        if response.content:
            return deserialize(response.json())
        else:
            return None

    def __build_endpoint(self, path):
        return os.path.join(self._config.api_base, path)

    def get(self, path: str, **kwargs):
        """
        Send get request and return retult
        """
        return self.__make_request("get", path, **kwargs)

    def post(self, path: str, data: Dict = None, **kwargs):
        """
        Send post request and return retult
        """
        data = json_dumps(serialize(data))
        return self.__make_request("post", path, data=data, **kwargs)

    def patch(self, path: str, data: Dict = None, **kwargs):
        """
        Send patch request and return retult
        """
        data = json_dumps(serialize(data))
        return self.__make_request("patch", path, data=data, **kwargs)

    def delete(self, path: str, **kwargs):
        """
        Send delete request and return retult
        """
        return self.__make_request("delete", path, **kwargs)
